# Import required libraries
import os
import glob

# Define atomic numbers to calculate multiplicity
ATOMIC_NUMBERS = {
    'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
    'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20,
    'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30,
    'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40,
    'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50,
    'Sb': 51, 'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60,
    'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70,
    'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80,
    'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86
}

# Invert the ATOMIC_NUMBERS dictionary to get atomic symbols
ATOMIC_SYMBOLS = {value: key for key, value in ATOMIC_NUMBERS.items()}

# Define the file types supported by the program
FILE_TYPES = {
    "xyz": {"skip_lines": 2, "split_func": lambda x: x.split() if x.strip() else []},
    "gjf": {"skip_func": lambda x: x.strip() == "0 1", "split_func": lambda x: x.split() if x.strip() and len(x.split()) == 4 else None},
    "pdb": {"split_func": lambda x: x.split()[2:6] if x[:4] == "ATOM" else None},
    "sdf": {"num_atoms_line": 3, "split_func": lambda x: x.split()[:4]},
    "mol": {"num_atoms_line": 3, "split_func": lambda x: x.split()[:4]},
    "mol2": {
        "start_func": lambda x: "@<TRIPOS>ATOM" in x,
        "end_func": lambda x: "@<TRIPOS>BOND" in x,
        "split_func": lambda x: [x.split()[1]] + x.split()[2:5] if len(x.split()) >= 5 else None,
        "offset": 1
    },
    "log": {
            "start_func": lambda x: "Standard orientation:" in x,
            "end_func": lambda x: "Rotational constants (GHZ):" in x,
            "split_func": lambda x: [ATOMIC_SYMBOLS[int(x.split()[1])]] + x.split()[3:6] if len(x.split()) == 6 and x.split()[1].isdigit() else None,
            "offset": 4
    }
}

# Define calculation type dictionary
CALC_TYPE_DICT = {
    "Single point": "",
    "Optimization": "opt",
    "Frequency": "freq tightSCF",
    "Optimization + Frequency": "opt freq tightSCF",
    "Molecular dynamics (MD)": "MD"
}

# Define method dictionary
METHOD_DICT = {
    "B97-3c": "! B97-3c{} noautostart miniprint nopop",
    "r2SCAN-3c": "! r2SCAN-3c{} noautostart miniprint nopop",
    "RI-B3LYP-D3(BJ)/def2-SVP": "! B3LYP D3 def2-SVP def2/J RIJCOSX{} noautostart miniprint nopop",
    "RI-B3LYP-D3(BJ)/def2-TZVP(-f)": "! B3LYP D3 def2-TZVP(-f) def2/J RIJCOSX{} noautostart miniprint nopop",
    "RI-B3LYP-D3(BJ)/def2-TZVP": "! B3LYP D3 def2-TZVP def2/J RIJCOSX{} noautostart miniprint nopop",
    "RI-PBE0-D3(BJ)/def2-TZVP": "! PBE0 D3 def2-TZVP def2/J RIJCOSX{} noautostart miniprint nopop",
    "RI-wB97X-D3/def2-TZVP": "! wB97X-D3 def2-TZVP def2/J RIJCOSX{} noautostart miniprint nopop",
    "RI-M06-2X/def2-TZVP": "! M06-2X def2-TZVP def2/J RIJCOSX{} defgrid3 noautostart miniprint nopop",
    "RI-wB97M-V/def2-TZVP": "! wB97M-V def2-TZVP def2/J RIJCOSX strongSCF{} noautostart miniprint nopop",
    "DLPNO-CCSD(T)/cc-pVTZ": "! DLPNO-CCSD(T) normalPNO RIJK cc-pVTZ cc-pVTZ/JK cc-pVTZ/C tightSCF{} noautostart miniprint nopop",
    "DLPNO-CCSD(T)/def2-TZVPP": "! DLPNO-CCSD(T) normalPNO RIJK def2-TZVPP def2/JK def2-TZVPP/C tightSCF{} noautostart miniprint nopop"
}

# Define MD commands for ORCA
MD_COMMANDS = '''%md
#restart ifexists  # Continue MD by reading [basename].mdrestart if it exists. In this case "initvel" should be commented
#minimize  # Do minimization prior to MD simulation
 timestep 0.5_fs  # This stepsize is safe at several hundreds of Kelvin
 initvel 100_K no_overwrite # Assign velocity according to temperature for atoms whose velocities are not available
 thermostat CSVR 298.15_K timecon 30.0_fs  # Target temperature and coupling time constant
 dump position stride 1 format xyz filename "trajectory.xyz"  # Dump position every "stride" steps
 run 2000 CenterCOM # Number of MD steps. Remove motion of center of mass
end'''

def read_file(file_name, file_type):
    """
    Read a file based on the file type.
    """
    with open(file_name, 'r') as f:
        if file_type == "gjf":
            lines = f.readlines()
            charge_multiplicity_line = [i for i, line in enumerate(lines) if line.strip() == '']
            charge, multiplicity = map(int, lines[charge_multiplicity_line[1] + 1].split())
            atom_lines = lines[charge_multiplicity_line[1] + 2:charge_multiplicity_line[2]]
            atoms = [line.split()[:4] for line in atom_lines if line.strip()]
            return atoms, charge, multiplicity
        else:
            try:
                if "skip_lines" in FILE_TYPES[file_type]:
                    lines = f.readlines()[FILE_TYPES[file_type]["skip_lines"]:]
                elif "skip_func" in FILE_TYPES[file_type]:
                    lines = f.readlines()
                    start = [i for i, line in enumerate(lines) if FILE_TYPES[file_type]["skip_func"](line)][0]
                    lines = lines[start+1:]
                elif "filter_func" in FILE_TYPES[file_type]:
                    lines = [line for line in f.readlines() if FILE_TYPES[file_type]["filter_func"](line)]
                elif "num_atoms_line" in FILE_TYPES[file_type]:
                    lines = f.readlines()
                    num_atoms = int(lines[FILE_TYPES[file_type]["num_atoms_line"]][:3].strip())
                    lines = lines[4:4+num_atoms]
                elif "start_func" in FILE_TYPES[file_type]:
                    lines = f.readlines()
                    start_indices = [i for i, line in enumerate(lines) if FILE_TYPES[file_type]["start_func"](line)]
                    end_indices = [i for i, line in enumerate(lines) if FILE_TYPES[file_type]["end_func"](line)]
                    start = start_indices[-1]  # Get the last 'Standard orientation:' index
                    end = end_indices[-1]  # Get the last 'Rotational constants (GHZ):' index
                    lines = [lines[i] for i in range(start + FILE_TYPES[file_type]["offset"], end)]
                else:
                    lines = f.readlines()
            except Exception as e:
                print(f"Error during line splitting in file {file_name}: {str(e)}")
                return []

            # For PDB files, properly parse the atom and coordinate lines
            if file_type == "pdb":
                atoms = []
                for line in lines:
                    if line.startswith(("ATOM", "HETATM")):  # Only consider ATOM and HETATM records
                        atom = line[12:16].strip()
                        x = line[30:38].strip()
                        y = line[38:46].strip()
                        z = line[46:54].strip()
                        atoms.append([atom, x, y, z])
            else:
                # Read lines until blank line is encountered
                atoms = [FILE_TYPES[file_type]["split_func"](line) for line in lines if line.strip()]
                atoms = [line for line in atoms if line]

            return atoms

def calculate_multiplicity(atoms):
    """
    Calculate the multiplicity of a set of atoms.
    """
    return 1 if sum(ATOMIC_NUMBERS.get(atom[0], 0) for atom in atoms if atom) % 2 == 0 else 2

def generate_orca_input(calc_type, method):
    """
    Generate ORCA input based on the calculation type and method.
    """
    calc_type_command = CALC_TYPE_DICT[calc_type]
    calc_type_command = f" {calc_type_command}" if calc_type_command else ""  # add a space in front if not empty
    method_command = METHOD_DICT[method].format(calc_type_command)

    if calc_type == "Molecular dynamics (MD)":
        method_command += "\n" + MD_COMMANDS

    return method_command

def convert_to_orca(input_file, output_file, calc_type, method, core=6, memory_per_core=4000):
    """
    Convert the input file (of various formats) into ORCA input file.
    """
    file_type = input_file.split('.')[-1]
    if file_type not in FILE_TYPES:
        print("Unsupported file format.")
        return

    if file_type == "gjf":
        atoms, charge, multiplicity = read_file(input_file, file_type)
    else:
        atoms = read_file(input_file, file_type)
        multiplicity = calculate_multiplicity(atoms)
        charge = 0  # Add default charge here

    print(f"Processing file: {input_file}")
    # The following line is commented out to prevent printing of the atom list
    # print(f"Atom list: {atoms}")

    with open(output_file, 'w') as f:
        f.write(generate_orca_input(calc_type, method) + "\n")
        f.write(f"%maxcore {memory_per_core}\n%pal nprocs {core} end\n")
        f.write("* xyz {} {}\n".format(charge, multiplicity))
        for atom in atoms:
            f.write("  ".join(atom) + "\n")
        f.write("*")

def main():
    """
    Main function to interact with the user and perform the conversion.
    """
    # Print welcome message
    print(
        "********************************************************************\n"
        "*                                                                  *\n"
        "*                       ORCA Input Generator                       *\n"
        "*                                                                  *\n"
        "*                          By Faizan Ullah                         *\n"
        "*                  Email: faizan@faizanullah.com                   *\n"
        "*                                                                  *\n"
        "*                      Release Date: 29 July, 2023                 *\n"
        "*                                                                  *\n"
        "********************************************************************\n"
        "\nWelcome to the ORCA Input Generator. This tool is designed to assist\n"
        "you in creating input files for ORCA, a quantum chemistry software package.\n"
        "Please follow the prompts to generate your input.\n"
    )

    try:
        print("Choose mode:")
        print("1. Single file mode")
        print("2. Batch mode")
        mode = input()

        if mode == "1":
            print("Enter the filename:")
            input_file = input()

        while True:
            try:
                core = input("Enter the number of cores (default is 6, enter an integer value): ")
                core = int(core) if core else 6
                break
            except ValueError:
                print("Invalid input. Please enter an integer value.")
        
        while True:
            try:
                memory_per_core = input("Enter the memory per core in MBs (default is 4000, enter an integer value): ")
                memory_per_core = int(memory_per_core) if memory_per_core else 4000
                break
            except ValueError:
                print("Invalid input. Please enter an integer value.")

        print("Choose calculation type:")
        calc_types = list(CALC_TYPE_DICT.keys())
        for i, calc_type in enumerate(calc_types, 1):
            print(f"{i}. {calc_type}")
        calc_type_index = int(input()) - 1  # convert from 1-based to 0-based index

        methods = list(METHOD_DICT.keys()) if calc_type_index == 0 else list(METHOD_DICT.keys())[:8]
        print("Choose method:")
        for i, method in enumerate(methods, 1):
            print(f"{i}. {method}")
        method_index = int(input()) - 1  # convert from 1-based to 0-based index

        if mode == "1":
            convert_to_orca(input_file, f"{os.path.splitext(input_file)[0]}.inp", calc_types[calc_type_index], methods[method_index], core, memory_per_core)
        else:
            for file in glob.glob("*"):
                if file.split('.')[-1] in FILE_TYPES:
                    convert_to_orca(file, f"{os.path.splitext(file)[0]}.inp", calc_types[calc_type_index], methods[method_index], core, memory_per_core)

        print("Operation completed successfully!")
    except Exception as e:
        print(f"An error occurred: {str(e)}")
    finally:
        input("Press Enter to exit...")

if __name__ == "__main__":
    main()